//--------------------------------------------
// NAME: Konstantin Ivanov
// CLASS: XIa
// NUMBER: 13
// PROBLEM: #2
// FILE NAME: list.hh
// FILE PURPOSE:
// implementation of the class List
//---------------------------------------------


#ifndef LIST_HH_
#define LIST_HH_

#include <exception>
template<class T>
class List {
	
	struct Elem {
		T data_;
		Elem* prev_;
		Elem* next_;
		
		Elem(const T& val)
		: data_(val),
		  prev_(0),
		  next_(0)
		{}
	};
	
	Elem* head_;
	Elem* tail_;

public:
	
	List() //constructor
	: head_(0),
	  tail_(head_)
	{}
	
	

	List(const List& other) //copy constructor
	: head_(0), tail_(0) {
		Elem* ptr=other.head_;
		while(ptr!=0) {
			push_back(ptr->data_);
			ptr=ptr->next_;
		}// returns the last element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	}
	
	~List() { // destructor
		clear();
	}
	
//--------------------------------------------
// FUNCTION: int size()
// returns the size of the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	int size() const {
		int count = 0;
		for( const_iterator it = begin(); it != end(); it++ )
			count++;
		return count;
	}
	
//--------------------------------------------
// FUNCTION: void clear()
// clears the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	void clear() {
		while(!empty()) {
			pop_back();		
		}
	}
	
//--------------------------------------------
// FUNCTION: bool empty()
// checks if the list is empty
// PARAMETERS:
// no parameters
//----------------------------------------------	
	bool empty() {
		return head_==0;
	}

//--------------------------------------------
// FUNCTION: void swap(List& other)
// swaps the contеnts of two lists
// PARAMETERS:
// List& other if the second list
//----------------------------------------------
	void swap(List& other) {
		Elem* tmp = head_;
		
		head_=other.head_;
		other.head_ = tmp;
		
		tmp = tail_;
		tail_=other.tail_;
		other.tail_=tmp;
	}
	
	
//--------------------------------------------
// FUNCTION: void push_back(const T& val)
// adds new element at the end of the list
// PARAMETERS:
// const T& val is the new element
//----------------------------------------------	
	void push_back(const T& val) {
		if(empty()) {
			head_=tail_=new Elem(val);
		} else {
			Elem* newElem = new Elem(val);
			
			tail_->next_=newElem;
			newElem->prev_=tail_;
			tail_=newElem;	
		}
	}

//--------------------------------------------
// FUNCTION: void pop_back()
// deletes the last element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	void pop_back() {
		if(empty()) {
			throw std::exception();
		} else if(head_==tail_) {
			delete head_;
			head_=tail_=0;	
		} else {
			Elem* tmp = tail_;
			
			tail_ = tail_->prev_;
			tail_->next_ = 0;
			
			delete tmp; 
		}
	}

//--------------------------------------------
// FUNCTION: void push_front(T val)
// adds new element at the beggining of the list
// PARAMETERS:
// T val is the new element
//----------------------------------------------
	void push_front(T val) {
		Elem* newElem = new Elem(val);
		if(empty()) {
			head_=tail_=newElem;
		} else {
			head_->prev_=newElem;
			newElem->next_=head_;
			head_=newElem;	
		}
	}

//--------------------------------------------
// FUNCTION: void pop_front()
// deletes the first element of the list
// PARAMETERS:
// no parameters
//----------------------------------------------	
	void pop_front() {
		if(empty()) {
			throw std::exception();
		} else if (head_==tail_) { //if there is only one element
			delete head_;
			head_=tail_=0;
		} else {				//if there are more than one element
			Elem* ptr=head_;
			head_=head_->next_;
			head_->prev_=0;
			delete ptr;
		}
		
	}

//--------------------------------------------
// FUNCTION: T& back()
// returns the last element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	T& back() {
		if(empty()) {
			throw std::exception();
		}
		return tail_->data_;
	}

//--------------------------------------------
// FUNCTION: const T& back()
// returns the last element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	const T& back() const {
		if(empty()) {
			throw std::exception();
		}
		return tail_->data_;		
	}

//--------------------------------------------
// FUNCTION: T& front()
// returns the first element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------
	T& front() {
		if(empty()) {
			throw std::exception();
		}
		return head_->data_;
	}

//--------------------------------------------
// FUNCTION: const T& front()
// returns the first element in the list
// PARAMETERS:
// no parameters
//----------------------------------------------	
	const T& front() const {
		if(empty()) {
			throw std::exception();
		}
		return head_->data_;		
	}

//--------------------------------------------
// FUNCTION: List& operator=(const List& other)
// predefinition of the operator=
// PARAMETERS:
// const List& other is the other list
//----------------------------------------------	
	List& operator=(const List& other) {
		if(this!=&other) {
			while(!empty()) {
				pop_back();
			}
			Elem* ptr=other.head_;
			while(ptr!=0) {
				push_back(ptr->data_);
				ptr=ptr->next_;
			}
		}
		return *this;
	}
	
	class iterator {
		Elem* curr_;
		
		iterator(Elem* e) 
		: curr_(e)
		{}
		friend class List;
	public:
		iterator operator++() { // preincrement
			curr_=curr_->next_;
			return *this;
		}
		
		iterator operator++(int) {  // postincrement
			iterator ret=*this;
			curr_=curr_->next_;
			return ret;
		}
		
//--------------------------------------------
// FUNCTION: bool operator==(const iterator& other)
// checks if two iterators are equal
// PARAMETERS:
// const iterator& other is the second iterator
//----------------------------------------------
		bool operator==(const iterator& other) {
			return curr_==other.curr_;
		}

//--------------------------------------------
// FUNCTION: bool operator!=(const iterator& other)
// checks if two iterators are not equal
// PARAMETERS:
// const iterator& other is the second iterator
//----------------------------------------------		
		bool operator!=(const iterator& other) {
			return !operator==(other);
		}

//--------------------------------------------
// FUNCTION: T& operator*()
// returns the value of the iterator
// PARAMETERS:
// no parameters
//----------------------------------------------		
		T& operator*() {
			return curr_->data_;
		}
		
		T* operator->() {            
               return &curr_->data;
          }
	}; 
	friend class iterator;

//--------------------------------------------
// FUNCTION: iterator begin()
// returns iterator directed to the first element
// PARAMETERS:
// no parameters
//----------------------------------------------
	iterator begin() {
		return iterator(head_);
	}

//--------------------------------------------
// FUNCTION: iterator begin()
// returns iterator directed to the last element
// PARAMETERS:
// no parameters
//----------------------------------------------
	iterator end() {
		return iterator(tail_);
	}

//--------------------------------------------
// FUNCTION: iterator insert(iterator pos, const T& x)
// inserts an element x before the element with 
// position pos
// PARAMETERS:
// pos is the position, x is the new element
//----------------------------------------------	
	iterator insert(iterator pos, const T& x) {
		Elem* tmp = pos.curr_;
		Elem* newElem = new Elem(x);
		if (pos == begin()) {
			push_front(x);
		} else if (pos == end()) {
			tail_->prev_->next_ = newElem;
			newElem->next_ = tail_;
			newElem->prev_ = tail_->prev_;
			tail_->prev_ = newElem;
		} else {
			tmp->prev_->next_ = newElem;
          	newElem->next_ = tmp;
          	newElem->prev_ = tmp->prev_;
          	tmp->prev_ = newElem;			
		}
	}

//--------------------------------------------
// FUNCTION: iterator erase(iterator pos)
// deletes the element with position pos
// PARAMETERS:
// iterator pos is the position 
//----------------------------------------------		
	iterator erase(iterator pos) {
		if (pos == begin()) {
			pop_front();
		} else if (pos == end()) {
			pop_back();
		} else {
			Elem* tmp = pos.curr_;
			tmp->prev_->next_ = tmp->next_;
			tmp->next_->prev_ = tmp->prev_;	
			delete tmp;	
		}
	}

//--------------------------------------------
// FUNCTION: iterator erase(iterator first, iterator last)
// deletes every element between first and last
// PARAMETERS:
// first is the first element that has to be deleted
// and last is the last one
//----------------------------------------------		
	iterator erase(iterator first, iterator last) {
		for(iterator it = first; it != last; it++) {
			erase(it);		
		}
	}
	
	class const_iterator {
		Elem* curr_;
		
		const_iterator(Elem* e) 
		: curr_(e)
		{}
		friend class List;
	public:
		const_iterator operator++(void) { // preincrement
			curr_=curr_->next_;
			return *this;
		}
		
		const_iterator operator++(int) {  // postincrement
			const_iterator ret=*this;
			curr_=curr_->next_;
			return ret;
		}
		
		bool operator==(const const_iterator& other) const {
			return curr_==other.curr_;
		}
		
		bool operator!=(const const_iterator& other) const {
			return !operator==(other);
		}
		
		T& operator*() {
			return curr_->data_;
		}
		
		T* operator->() {            
               return &curr_->data;
          }
	}; 
	friend class const_iterator;
	
	const_iterator begin() const {
		return const_iterator(head_);
	}
	
	const_iterator end() const {
		return const_iterator(tail_);
	}
	
	class reverse_iterator {
		Elem* curr_;
		
		reverse_iterator(Elem* e) 
		: curr_(e)
		{}
		friend class List;
	public:
		reverse_iterator operator++(void) { // preincrement
			curr_=curr_->prev_;
			return *this;
		}
		
		reverse_iterator operator++(int) {  // postincrement
			reverse_iterator ret=*this;
			curr_=curr_->prev_;
			return ret;
		}
		
		bool operator==(const reverse_iterator& other) {
			return curr_==other.curr_;
		}
		
		bool operator!=(const reverse_iterator& other) {
			return !operator==(other);
		}
		
		T& operator*()  {
			return curr_->data_;
		}
		
		T* operator->() {            
               return &curr_->data;
          }
	}; 
	friend class reverse_iterator;
	
	reverse_iterator rbegin() {
		return reverse_iterator(tail_);
	}
	
	reverse_iterator rend() {
		return reverse_iterator(0);
	}
	
	class const_reverse_iterator {
		Elem* curr_;
		
		const_reverse_iterator(Elem* e) 
		: curr_(e)
		{}
		friend class List;
	public:
		const_reverse_iterator operator++(void) { // preincrement
			curr_=curr_->prev_;
			return *this;
		}
		
		const_reverse_iterator operator++(int) {  // postincrement
			const_reverse_iterator ret=*this;
			curr_=curr_->prev_;
			return ret;
		}
		
		bool operator==(const const_reverse_iterator& other) const {
			return curr_==other.curr_;
		}
		
		bool operator!=(const const_reverse_iterator& other) const {
			return !operator==(other);
		}
		
		T& operator*() {
			return curr_->data_;
		}
		
		T* operator->() {            
               return &curr_->data;
          }
	}; 
	friend class const_reverse_iterator;
	
	const_reverse_iterator rbegin() const {
		return const_reverse_iterator(tail_);
	}
	
	const_reverse_iterator rend() const {
		return const_reverse_iterator(0);
	}
	
	

};

#endif


















